 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */
package Java2OWL;

import java.lang.reflect.Method;
import java.util.*;
import java.util.Map.Entry;
import org.semanticweb.owlapi.model.*;

/** This class manages all correspondences between Java-classes and OWL-classes.
 * <p>For OWL-classes generated from Java-classes there is a one-to-one correspondence. 
 * In contrast to Java classes, OWL-classes allow for multiple inheritance.
 * This means that for subclasses there is no one-to-one correspondence any more. 
 * <p>
 * Example:
 * Java-classes Ship and LandVehicle are mapped to OWL-classes.
 * If in the extension ontology a common subclass AmphibiousCraft is externally added, this class
 * corresponds to the two Java-classes Ship and LandVehicle.
 * This becomes relevant when OWL-individuals are imported and mapped to Java-objects. 
 * In the case of AmphibiousCraft-instances, one could then generate two Java-objects,
 * a Ship and a LandVehicle. 
 * 
 * @author ohlbach 
 */
public class J2OClassManager {
    
    /** the top manager. */ 
    private J2OManager manager; 
    
    /** returns the name of the class manager.*/
    public String getName() {return manager.getName() + "_ClassManager";}
   
    /** The ontology manager to be used for the mapping between Java and OWL. */
    private J2OOntologyManager ontologyManager = null;

    /** maps Java-classes to ClassWrapper. */
    private Map<Class,ClassWrapper> directJ2Omap = new HashMap<Class,ClassWrapper>();
    
    /** returns for a given Java class the corresponding class wrapper, 
     * or null if there is none stored.*/
    public ClassWrapper getCLW(Class jcl) {return directJ2Omap.get(jcl);}
    
    /** maps OWL-classes to ClassWrapper. */
    private Map<OWLClass,ClassWrapper> directO2Jmap = new HashMap<OWLClass,ClassWrapper>();
    
     /** maps Java-classes and subclasses to ClassWrapper. 
     * <p> This map may be changed at run time and must therefore be thread save.*/
    private final Map<Class,ClassWrapper> cachedJ2Omap = new HashMap<Class,ClassWrapper>();
    
    /** returns the cached class wrapper for the given Java class .*/
    private ClassWrapper getCachedCLW(Class jcl) {
        synchronized(cachedJ2Omap) {return cachedJ2Omap.get(jcl);}}
    
     /** puts the pair [jcl,cls] into cachedJ2Omap. */
    private void putCachedCLW(Class jcl, ClassWrapper clw) {
        synchronized(cachedJ2Omap) {cachedJ2Omap.put(jcl,clw);}}
    
    /** maps OWL-classes and subclasses to ClassWrapper.
     * <p>Since OWL-class may have multiple direct superclasses there is no one-to-one relation.
     * <p> This map may be changed at run time and must therefore be thread save. */
    private final Map<OWLClass,Set<ClassWrapper>> cachedO2Jmap = new HashMap<OWLClass,Set<ClassWrapper>>();
    
    /** returns the cached class wrappers for the given owl class. */
    private Set<ClassWrapper> getCachedCLWs(OWLClass ocl) {
        synchronized(cachedO2Jmap) {return cachedO2Jmap.get(ocl);}}
    
    /** puts the pair [ocl,clws] into cachedO2Jmap. */
    private void putCachedCLWs(OWLClass ocl,Set<ClassWrapper> clws) {
        synchronized(cachedO2Jmap) {cachedO2Jmap.put(ocl,clws);}}
    
     /** puts the pair [ocl,{clw}] into cachedO2Jmap. */
    private void putCachedCLWs(OWLClass ocl,ClassWrapper clw) {
        Set<ClassWrapper> clws = new HashSet<ClassWrapper>();
        clws.add(clw);
        synchronized(cachedO2Jmap) {cachedO2Jmap.put(ocl,clws);}}
    
    
    
    /** maps getter-methods to PropertyWrapper. */
    private Map<Method,PropertyWrapper> getterMap = new HashMap<Method,PropertyWrapper>();
    
    /** maps OWL-property-names to PropertyWrapper. */
    private Map<String,PropertyWrapper> propertyMap = new HashMap<String,PropertyWrapper>();
    
    /** logs events and error messages. */
    private Logging logger;
    
    /** creates a class manager for the given top-manager. */
    public J2OClassManager(J2OManager manager) {
        this.manager = manager;
        this.ontologyManager = manager.getOntologyManager();
        logger = manager.getLogger();
        logger.installation(this);}

   /** an empty constructor (for test purposes) */
   J2OClassManager() {}

    
    /** adds the correspondence Java-class <-> OWL class. 
     * <p>The mapping must be injective.
     * 
     * @param jcl the Java-class.
     * @param ocl the OWL-class.
     * @return a ClassWrapper-object for the new correspondence or null if this is not possible, for example if another class is already mapped to the OWL-class. 
     */
    ClassWrapper addClassCorrespondence(Class jcl, OWLClass ocl)  {
        ClassWrapper clw = directJ2Omap.get(jcl);
        if(clw != null && jcl == clw.getJavaClass()) return clw;
        logger.entering(jcl,ocl);
        Set<ClassWrapper> clws = getCachedCLWs(ocl);
        if(clws != null && !clws.isEmpty()) {
            clw = clws.iterator().next();
            logger.error("Another class is already mapped to OWL ", Logging.Origin.JAVA, jcl,clw.getJavaClass(),ocl);
            return null;}
        clw = new ClassWrapper(jcl,ocl,manager);
        directJ2Omap.put(jcl, clw);
        directO2Jmap.put(ocl, clw);    
        putCachedCLW(jcl, clw);
        putCachedCLWs(ocl, clw);
        logger.correlation(Logging.Semantics.OCC,jcl,ocl);
        logger.exiting(jcl,ocl);
        return clw;}
    
    /** returns the set of classes managed by the class manager. */
    public Set<Class> getClasses() {return directJ2Omap.keySet(); }
    
    /** returns the ClassWrappers managed by the class manager. */
    public Collection<ClassWrapper> getCLWs() {return directJ2Omap.values(); }
   
    /** returns all the PropertyWrappers for the classes. */
    Collection<PropertyWrapper> getPRWs() {return getterMap.values();}
    
    /** returns true if the Java class is directly mapped to an OWLClass. */
    public boolean isMapped(Class jcl) {return directJ2Omap.get(jcl) != null;}
    
    /** returns the ClassWrapper for the given Java class, or null if none can be found..
     * 
     * @param jcl a Java class.
     * @param direct if false then jcl's superclasses are also searched for ClassWrapper.
     * @param cache if true the finding is internally stored. 
     * @return the corresponding ClassWrapper, or null if none could be found.
     * <p>
     * cache = true should only be used if the Java <-> OWL relations do not change any more
     * for jcl's superclasses.
     */
    ClassWrapper getCLW(Class jcl, boolean direct, boolean cache) {
        if(direct) return directJ2Omap.get(jcl);
        ClassWrapper clw = getCachedCLW(jcl);
        if(clw != null) return clw;
        Class sjcl = jcl.getSuperclass();
        while(clw == null && sjcl != null) {
              clw = directJ2Omap.get(sjcl);
              sjcl = sjcl.getSuperclass();}
        if(cache)putCachedCLW(jcl,clw);
        return clw;}
    
    /** returns the ClassWrapper for the given Java class, or null if none can be found..
     * 
     * @param jcl a Java class.
     * @param direct if false then jcl's superclasses are also searched for ClassWrapper.
     * @return the corresponding ClassWrapper, or null if none could be found.
     */
     public ClassWrapper getCLW(Class jcl, boolean direct) {return getCLW(jcl,direct,true);}
    
    /** yields the OWL-class which corresponds to the Java class jcl, or null.
     * 
     * @param jcl the Java Class.
     * @param direct if false then jcl's superclasses are also searched for ClassWrapper.
     * @param cache if true the finding is internally stored. 
     * @return the corresponding OWL-Class.
     * <p>
     * cache = true should only be used if the Java <-> OWL relations do not change any more
     * for jcl's superclasses.
     */
    OWLClass getOWLClass(Class jcl, boolean direct, boolean cache) {
        ClassWrapper cmd = getCLW(jcl,direct,cache);
        return (cmd == null) ? null : cmd.getOWLClass();}
    
    
     /** yields the OWL-class which corresponds to the Java class jcl, or null.
     * 
     * @param jcl the Java Class.
     * @param direct if false then jcl's superclasses are also searched for ClassWrapper.
     * @param cache if true the finding is internally stored. 
     * @return the corresponding OWL-Class.
     */
    public OWLClass getOWLClass(Class jcl, boolean direct) {
        ClassWrapper cmd = getCLW(jcl,direct,true);
        return (cmd == null) ? null : cmd.getOWLClass();}
      
    /** yields the Java-classes which corresponds to the given OWL-class.
     * 
     * <p>
     * Since OWL has multiple inheritance, an OWL-class may have several superclasses.
     * which correspond to Java-classes. The lowest such superclasses are returned.
     * 
     * @param ocl The OWL-class.
     * @return the corresponding Java-classes.
     */
    Set<Class> getJavaClasses(OWLClass ocl) {
        Set<Class> classes = new HashSet<Class>();
        for(ClassWrapper  cmd : getCLWs(ocl)) classes.add(cmd.getJavaClass());
         return classes;}
     
    /** yields the ClassWrappers which corresponds to the given OWL-class.
     * 
     * <p>
     * Since OWL has multiple inheritance, an OWL-class may have several superclasses.
     * which correspond to Java-classes. The lowest such superclasses are returned.
     * <p> The returned ClassWrappers are such that their Java classes are not subclasses
     * of each other. Only the more specific subclasses are included.
     * 
     * @param ocl The OWL-class.
     * @return the corresponding set of ClassWrapper.
     */
    
    Set<ClassWrapper> getCLWs(OWLClass ocl) { 
        Set<ClassWrapper> clws = getCachedCLWs(ocl);
        if(clws != null) return clws;
        
        clws = new HashSet<ClassWrapper>();
        Set<OWLClass> ocls = new HashSet<OWLClass>();
        ocls.add(ocl);
        
        while(!ocls.isEmpty()) { 
            OWLClass owlcl = ocls.iterator().next();
            ocls.remove(owlcl); 
            ClassWrapper clwr = directO2Jmap.get(owlcl);
            if(clwr == null) {
                Set<OWLClass> superclasses = ontologyManager.getSuperclasses(owlcl, true);
                if(superclasses != null) ocls.addAll(superclasses);}
            else { // now we make sure that only the most specific Java classes are included.
                boolean moreSpecific = false;
                for(ClassWrapper oldclw : clws) {
                    if(oldclw.getJavaClass().isAssignableFrom(clwr.getJavaClass())) {
                        clws.remove(oldclw); break;} // new class is nore specific
                     if(clwr.getJavaClass().isAssignableFrom(oldclw.getJavaClass())) {
                        moreSpecific = true; break;}}// old class is more specific. New class is not needed.
                if(!moreSpecific) clws.add(clwr);}}
       
        putCachedCLWs(ocl,clws);
        return clws;}
 
    /** yields the ClassWrappers which correspond to the given OWL-classes.
     * 
     * <p>
     * Since OWL has multiple inheritance, an OWL-class may have several superclasses
     * which correspond to Java-classes. The lowest such superclasses are returned.
     * remember = true should only be used if the Java <-> OWL relations do not change any more
     * for ocl's superclasses.
     * 
     * @param ocls a set of OWL classes.
     * @return the corresponding set of ClassWrapper.
     */
    public Set<ClassWrapper> getCLWs(Set<OWLClass> ocls) { 
         Set<ClassWrapper> cmds = new HashSet<ClassWrapper>();
         for(OWLClass ocl : ocls) cmds.addAll(getCLWs(ocl));
         return cmds;}
    
    
   /** adds a getter-method to the class manager.
     * 
     * <p>
     * Either a new PropertyWrapper is created for the method, or an existing one is extended.
     * The OWL-Property is not yet created.
     * 
     * @param getter a getter method.
     * @throws J2OException if the analysis of the getter method yields an error.
     */
   PropertyWrapper addMethod(Class jcl, Method getter) throws J2OException {
       logger.entering(jcl,getter);
       String pname = PropertyWrapper.getPropertyName(getter); 
       PropertyWrapper pmd = propertyMap.get(pname);
       if(pmd == null) {
           pmd = PropertyWrapperFactory.makePropertyWrapper(jcl,getter,manager);
           propertyMap.put(pname,pmd);}
       else pmd.join(jcl,getter);
       getterMap.put(getter, pmd);
       
       ClassWrapper clm = directJ2Omap.get(jcl);
       assert(clm != null) : "The class " + jcl.getName() + " has not yet been compiled";
       clm.addProperty(pmd); 
       logger.correlation(Logging.Semantics.OPC, getter, pname);
       logger.exiting(jcl,getter);
       return pmd;}
   
   /** returns the PropertyWrapper for the given getter-method. */
   PropertyWrapper getPRW(Method getter) {return getterMap.get(getter);}
   
   
   
    /** activates Java <-> OWL synchronisation for all objects in the given classes.
    * 
    * @param classes are either single Java class, an array or a collection of Java classes. 
    * <p>For all these classes and their subclasses the synchronisation is activated.
    * This activation can be overwritten at object level. 
    * <p>Before the synchronisation is activated all objects in the classes are synchronised with OWL.
    * 
    * @see IndividualWrapper.activateSynchronisation.
    * @throw ClassCastException if the classes are no Java classes.
    */
    public void activateSynchronisation(Object... classes) {
        final J2OIndividualManager img = manager.getIndvidualManager();
        for(Object object : classes) {
            Utilities.map(object,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        Class jcl = (Class)object;
                        if(!J2OSynchroniser.isSynchronised(jcl)) return true;
                        for(Class subclass : getClasses()) {
                            if(jcl != subclass && jcl.isAssignableFrom(subclass))  activateSynchronisation(subclass);}
                        if(img != null) {
                            Set<IndividualWrapper> iwrs = img.getIWRs(jcl);
                            if(iwrs != null) {
                                for(IndividualWrapper iwr : iwrs) {iwr.synchroniseWithOWL();}}}
                        logger.info("synchronisation activated for class ", jcl);
                        try{Utilities.setStaticField(jcl, J2OSynchroniser.INDIVIDUAL_MANAGER_VARIABLE, img);}
                        catch(Exception ex){
                            logger.error("Field access error", Logging.Origin.JAVA, J2OSynchroniser.INDIVIDUAL_MANAGER_VARIABLE, ex.getMessage());}
                        return true;}});}}
    
    
    /** deactivates Java <-> OWL synchronisation for all objects in the given classes.
    * 
    * @param classes are either single Java class, an array or a collection of Java classes. 
    * <p>For all these classes and their subclasses the synchronisation is activated.
    * This activation can be overwritten at object level. 
     * 
    * @see IndividualWrapper.activateSynchronisation.
    * @throw ClassCastException if the classes are no Java classes.
    */
    public void deactivateSynchronisation(Object... classes) {
        for(Object object : classes) {
            Utilities.map(object,
                new Function() {
                    @Override
                    public boolean apply(Object object) {
                        Class jcl = (Class)object;
                        if(!J2OSynchroniser.isSynchronised(jcl)) return true;
                        for(Class subclass : getClasses()) {
                            if(jcl != subclass && jcl.isAssignableFrom(subclass))  activateSynchronisation(subclass);}
                        logger.info("synchronisation activated for class ", jcl);
                        try{Utilities.setStaticField(jcl, J2OSynchroniser.INDIVIDUAL_MANAGER_VARIABLE, null);}
                        catch(Exception ex){
                            logger.error("Field access error", Logging.Origin.JAVA, J2OSynchroniser.INDIVIDUAL_MANAGER_VARIABLE, ex.getMessage());}
                        return true;}});}}
    
    
   
   
   /** clears the internal cache. */
   public void clearCache() {
       synchronized(cachedO2Jmap){cachedO2Jmap.clear();}
       synchronized(cachedJ2Omap){cachedJ2Omap.clear();}}
   
   /** clears all internal class-relevant data. */
   public void clear() {
       clearCache();
       directJ2Omap.clear();
       directO2Jmap.clear();
       getterMap.clear();
       propertyMap.clear();}
   
   /** returns a string-representation of the class manager. */
    @Override
    public String toString() {
       String blanks = "   ";
       String dblanks = blanks + blanks;
       String s = "J2OClassManager " + getName() + "\n";
       if(ontologyManager != null) s += blanks + "Ontology Manager: " + getName() + "\n";
 
       s += blanks + "Direct Java -> ClassMetadata map:\n";
       for(Entry<Class,ClassWrapper> entry : directJ2Omap.entrySet()) 
           s += dblanks + entry.getKey().getSimpleName() + " -> " + entry.getValue().oClassShortName() + "\n";
  
       s += blanks + "Inherited Java -> ClassMetadata map:\n";
       for(Entry<Class,ClassWrapper> entry : cachedJ2Omap.entrySet()) {
           s += dblanks + entry.getKey().getSimpleName() + " -> ";
           if(entry.getValue() == null) s += "null";
           else s += entry.getValue().oClassShortName();
           s+= "\n";}
 
       s += blanks + "Direct OWLClass -> ClassMetadata map:\n";
       for(Entry<OWLClass,ClassWrapper> entry : directO2Jmap.entrySet()) 
           s += dblanks + entry.getKey().getIRI().getFragment() + " -> " + entry.getValue().jClassShortName() + "\n";
 
       s += blanks + "Inherited OWLClass -> ClassMetadata map:\n";
       for(Entry<OWLClass,Set<ClassWrapper>> entry : cachedO2Jmap.entrySet()) {
           s += dblanks + entry.getKey().getIRI().getFragment() + " -> \n";
           for(ClassWrapper cmd : entry.getValue()) {
               s += dblanks + blanks + cmd.jClassShortName() +"\n";}}
           
       s += blanks + "Method -> OWLProperty map: \n";
       for(Entry<Method,PropertyWrapper>  entry : getterMap.entrySet()) {
           s += dblanks + entry.getKey().getName() + "-> " + entry.getValue().getName() + "\n";
       }
        
       s += blanks + "Method names -> OWLProperty map: \n";
       for(Entry<String,PropertyWrapper>  entry : propertyMap.entrySet()) {
           s += dblanks + entry.getKey() + "-> " + entry.getValue().getName() + "\n";
       }     
       return s;}
}
